# Copyright 2026 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for gomoku."""

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

import pyspiel

FLAGS = flags.FLAGS


class GamesGomokuTest(parameterized.TestCase):

  def test_gomoku_game_funs(self):
    # Default 15 x 15 grid, 5 in a row wins, no wrap.
    game = pyspiel.load_game("gomoku")
    dims = game.dims()
    self.assertEqual(dims, 2, f"Incorrect dims {dims}")
    size = game.size()
    self.assertEqual(size, 15, f"Incorrect size {size}")
    connect = game.connect()
    self.assertEqual(connect, 5, f"Incorrect connect {connect}")
    wrap = game.wrap()
    self.assertEqual(wrap, False, f"Incorrect wrap {wrap}")
    anti = game.anti()
    self.assertEqual(anti, False, f"Incorrect anti {anti}")

    coord = [2, 3]
    action = game.move_to_action(coord=coord)
    print("action", action)
    move = game.action_to_move(action)
    self.assertEqual(coord, move, f"Coord {coord} move {move}")

  def test_gomoku_hash(self):
    game = pyspiel.load_game("gomoku(size=3,connect=3)")
    state = game.new_initial_state()
    hash0 = state.hash_value()
    self.assertEqual(hash0, 0, f"Initial board hash {hash0}")
    state.apply_action(1)
    hash1 = state.hash_value()
    sym1 = state.symmetric_hash()
    # States related by symmetry shoud have different hashes
    # but the same symmetric hash.
    state = game.new_initial_state()
    state.apply_action(5)
    hash2 = state.hash_value()
    sym2 = state.symmetric_hash()
    self.assertNotEqual(hash1, hash2, f"Hash1 {hash1} Hash2 {hash2}")
    self.assertEqual(sym1, sym2, f"Hash1 {sym1} Hash2 {sym2}")

    # We can change symmetry policy during a game.
    # Change allow_reflections from false to true.
    state = game.new_initial_state()
    policy = state.get_symmetry_policy()
    print("policy", policy)
    self.assertEqual(policy.allow_reflections, False, "Wrong symmetry policy")
    # verify policy is correct here
    state.apply_action(0)
    state.apply_action(1)
    sym11 = state.symmetric_hash()
    # set symmetry policy here
    policy.allow_reflections = True
    policy = state.get_symmetry_policy()
    self.assertEqual(policy.allow_reflections, True, "Wrong symmetry policy")
    sym12 = state.symmetric_hash()

    state = game.new_initial_state()
    policy = state.get_symmetry_policy()
    # verify policy is correct here
    self.assertEqual(policy.allow_reflections, False, "Wrong symmetry policy")
    state.apply_action(2)
    state.apply_action(1)
    sym21 = state.symmetric_hash()
    # set symmetry policy here
    policy.allow_reflections = True
    sym22 = state.symmetric_hash()
    self.assertNotEqual(sym11, sym21, f"Hash1 {sym11} Hash2 {sym21}")
    self.assertEqual(sym12, sym22, f"Hash1 {sym12} Hash2 {sym22}")

  def test_gommoku_game_sim(self):
    game = pyspiel.load_game("gomoku")
    for _ in range(10):
      state = game.new_initial_state()
      mc = 0
      while not state.is_terminal():
        legal_actions = state.legal_actions()
        action = np.random.choice(legal_actions)
        state.apply_action(action)
        mc += 1

  def test_winning_line(self):
    game = pyspiel.load_game("gomoku(size=3,connect=3)")
    state = game.new_initial_state()
    state.apply_action(0)
    state.apply_action(1)
    state.apply_action(4)
    state.apply_action(2)
    state.apply_action(8)
    win = state.winning_line()
    self.assertEqual(win, [[0, 0], [1, 1], [2, 2]], f"win {win}")
    returns = state.returns()
    self.assertEqual(returns, [1.0, -1.0], f"returns {returns}")

    # same "winning line", but with anti it is the losing line.
    game = pyspiel.load_game("gomoku(size=3,connect=3,anti=true)")
    state = game.new_initial_state()
    state.apply_action(0)
    state.apply_action(1)
    state.apply_action(4)
    state.apply_action(2)
    state.apply_action(8)
    win = state.winning_line()
    self.assertEqual(win, [[0, 0], [1, 1], [2, 2]], f"win {win}")
    returns = state.returns()
    self.assertEqual(returns, [-1.0, 1.0], f"returns {returns}")

    # wrap winning line
    game = pyspiel.load_game("gomoku(size=4,connect=3,wrap=true)")
    state = game.new_initial_state()
    state.apply_action(2)
    state.apply_action(3)
    state.apply_action(7)
    state.apply_action(6)
    state.apply_action(8)
    win = state.winning_line()
    self.assertEqual(win, [[0, 2], [1, 3], [2, 0]], f"win {win}")
    returns = state.returns()
    self.assertEqual(returns, [1.0, -1.0], f"returns {returns}")

  def test_consistent_hash(self):
    game = pyspiel.load_game("gomoku")
    state = game.new_initial_state()
    state.apply_action(0)
    hash1 = state.hash_value()
    game = pyspiel.load_game("gomoku")
    state = game.new_initial_state()
    state.apply_action(0)
    hash2 = state.hash_value()
    self.assertEqual(hash1, hash2, f"Hash1 {hash1} Hash2 {hash2}")


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN agent implemented in PyTorch."""

import enum
import functools
import pathlib
from typing import Callable, Iterable, NamedTuple

import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
import tree as np_tree

from open_spiel.python import rl_agent
from open_spiel.python import rl_environment

ILLEGAL_ACTION_LOGITS_PENALTY = torch.finfo(torch.float).min


class Transition(NamedTuple):
  """Data structure for the Replay buffer."""

  info_state: np.ndarray
  action: np.ndarray
  reward: np.ndarray
  next_info_state: np.ndarray
  is_final_step: np.ndarray
  legal_actions_mask: np.ndarray


class ReplayBuffer:
  """Generic Replay memory for DQN."""

  def __init__(self, capacity: np.ndarray, experience: Transition) -> None:
    """Replay buffer constructor.

    Args:
      capacity (np.ndarray): capacity for the buffer.
      experience (Transition): an initialising sample for the buffer.
    """
    self.capacity = capacity
    self.experience = experience
    self.entry_index = np.array(0)

  def __len__(self) -> int:
    """Returns length of the buffer."""
    return min(self.entry_index, self.capacity).astype(int)

  @classmethod
  def init(cls, capacity: int, experience: Transition) -> "ReplayBuffer":
    # Initialize buffer by replicating the structure of the experience
    experience_ = np_tree.map_structure(
        lambda x: np.empty((capacity, *x.shape), dtype=x.dtype), experience
    )
    return cls(np.array(capacity), experience_)

  def append(
      self,
      experience: Transition,
  ) -> None:
    """Potentially adds `experience` to the replay buffer.

    Args:
      experience: data to be added to the replay buffer.

    Returns:
      None as the method updated the buffer in-place.
    """
    index = self.entry_index % self.capacity

    def _inplace(arr, idx, val):
      arr[idx] = val

    np_tree.map_structure(
        lambda buf_leaf, exp_leaf: _inplace(buf_leaf, index, exp_leaf),
        self.experience,
        experience,
    )

    self.entry_index += 1

  def sample(self, num_samples: int) -> Transition:
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.

    Raises:
      ValueError: If there are less than `num_samples` elements in the buffer.
    """
    max_size = len(self)
    if max_size < num_samples:
      raise ValueError(
          f"{num_samples} elements could not be sampled from size {max_size}"
      )

    indices = np.random.choice(max_size, size=(num_samples,), replace=False)

    return np_tree.map_structure(lambda data: data[indices], self.experience)


def set_seed(seed):
  np.random.seed(seed)
  torch.manual_seed(seed)
  if torch.cuda.is_available():
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


class MLP(nn.Module):
  """A simple network built from nn.linear layers."""

  def __init__(
      self,
      input_size: int,
      hidden_sizes: Iterable[int],
      output_size: int,
      final_activation: nn.Module = None,
      seed: int = 42,
  ) -> None:
    """Create the MLP.

    Args:
      input_size: (int) number of inputs.
      hidden_sizes: (list) sizes (number of units) of each hidden layer.
      output_size: (int) number of outputs.
      final_activation: (nn.Module) final activation of the network. Defaults to
        None.
      seed: (int) seed for the random number generator.
    """
    super().__init__()
    set_seed(seed)
    layers_ = []

    def _create_linear_block(in_features, out_features):
      return nn.Sequential(nn.Linear(in_features, out_features), nn.ReLU())

    # Input and Hidden layers
    for size in hidden_sizes:
      layers_.append(_create_linear_block(input_size, size))
      input_size = size
    # Output layer
    layers_.append(nn.Linear(input_size, output_size))
    if final_activation:
      layers_.append(final_activation)
    self.model = nn.Sequential(*layers_)

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    return self.model(x)


class Loss(enum.StrEnum):
  MSE = "mse"
  HUBER = "huber"


class Optimiser(enum.StrEnum):
  SGD = "sgd"
  RMSPROP = "rmsprop"
  ADAM = "adam"


class EpsilonDecaySchedule(enum.StrEnum):
  LINEAR = "linear"
  EXP = "exp"


# EPSILON DECAY SCHEDULES
def exponential_schedule(
    start_e: float, end_e: float, duration: float
) -> Callable[[int], float]:
  def _call(t: int) -> float:
    decay_steps = min(t, duration)
    return end_e + (start_e - end_e) * np.exp(-1.0 * decay_steps / duration)

  return _call


def linear_schedule(start_e: float,
                    end_e: float,
                    duration: int) -> Callable[[int], float]:
  slope = (end_e - start_e) / duration

  def _call(t: int) -> float:
    return max(slope * t + start_e, end_e)

  return _call


# pylint: disable=g-bare-generic
class DQN(rl_agent.AbstractAgent):
  """DQN Agent implementation in PyTorch.

  See open_spiel/python/examples/breakthrough_dqn.py for an usage example.
  """

  def __init__(
      self,
      player_id: int,
      state_representation_size: tuple[int, ...],
      num_actions: int,
      hidden_layers_sizes: Iterable[int] = (128,),
      batch_size: int = 128,
      replay_buffer_class=ReplayBuffer,
      replay_buffer_capacity: int = 10000,
      learning_rate: float = 0.01,
      update_target_network_every: int = 1000,
      weight_update_coeff: float = 0.995,
      learn_every: int = 10,
      discount_factor: float = 1.0,
      min_buffer_size_to_learn: int = 1000,
      epsilon_start: float = 1.0,
      epsilon_end: float = 0.1,
      epsilon_decay_duration: int = int(1e6),
      epsilon_decay_schedule_str: EpsilonDecaySchedule = "exp",
      optimizer_str: Optimiser = "sgd",
      loss_str: Loss = "mse",
      huber_loss_parameter: float = 1.0,
      seed: int = 42,
      gradient_clipping: float | None = None,
      device: str = "cpu",
  ) -> None:
    """Initialize the DQN agent."""
    # This call to locals() is used to store every argument used to initialise
    # the class instance, so it can be copied with no hyperparameter change.
    self._kwargs = locals()
    # Some type checks
    assert isinstance(player_id, int | None)
    assert isinstance(seed, int | None)
    assert isinstance(num_actions, int)
    assert isinstance(batch_size, int)
    assert isinstance(replay_buffer_capacity, int)
    assert isinstance(learn_every, int)
    assert isinstance(min_buffer_size_to_learn, int)
    assert isinstance(discount_factor, float)

    self.player_id = player_id
    self._num_actions = num_actions
    if isinstance(hidden_layers_sizes, int):
      hidden_layers_sizes = [hidden_layers_sizes]
    self._layer_sizes = hidden_layers_sizes
    self._batch_size = batch_size
    self._update_target_network_every = update_target_network_every
    self._learn_every = learn_every
    self._min_buffer_size_to_learn = min_buffer_size_to_learn
    self._discount_factor = discount_factor
    assert discount_factor >= 0 and discount_factor <= 1

    self._device = torch.device(device)

    self._replay_buffer_class = replay_buffer_class

    assert hasattr(replay_buffer_class, "init")
    assert hasattr(replay_buffer_class, "append")
    assert hasattr(replay_buffer_class, "sample")

    self._replay_buffer_capacity = int(replay_buffer_capacity)
    self._replay_buffer = None

    self._tau = weight_update_coeff
    assert self._tau >= 0 and self._tau < 1

    self._prev_timestep = None
    self._prev_action = None

    # Step counter to keep track of learning, eps decay and target network.
    self._iteration = 0

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    # Create the Q-network instances
    self._q_network = MLP(
        state_representation_size,
        self._layer_sizes,
        num_actions,
        None,  # outputs = raw logits
        seed=seed,
    ).to(self._device)

    self._target_q_network = MLP(
        state_representation_size,
        self._layer_sizes,
        num_actions,
        None,  # outputs = raw logits
        seed=seed,
    ).to(self._device)

    self._target_q_network.load_state_dict(self._q_network.state_dict())

    if loss_str == Loss.MSE:
      self.loss_class = F.mse_loss
    elif loss_str == Loss.HUBER:
      self.loss_class = functools.partial(
          F.smooth_l1_loss, beta=huber_loss_parameter
      )
    else:
      raise ValueError("Not implemented, choose from 'mse', 'huber'.")

    assert (epsilon_start > 0 and epsilon_end >= 0) and (
        epsilon_start >= epsilon_end
    )
    if epsilon_decay_schedule_str == EpsilonDecaySchedule.EXP:
      self.epsilon_schedule = exponential_schedule(
          epsilon_start, epsilon_end, epsilon_decay_duration
      )
    elif epsilon_decay_schedule_str == EpsilonDecaySchedule.LINEAR:
      self.epsilon_schedule = linear_schedule(
          epsilon_start, epsilon_end, epsilon_decay_duration
      )
    else:
      raise ValueError("Not implemented, choose from 'linear', 'exp'.")

    self._gradient_norm_clipping = gradient_clipping

    if optimizer_str == Optimiser.ADAM:
      self._optimizer = torch.optim.Adam(
          self._q_network.parameters(), lr=learning_rate
      )
    elif optimizer_str == Optimiser.RMSPROP:
      self._optimizer = torch.optim.RMSprop(
          self._q_network.parameters(), lr=learning_rate
      )
    elif optimizer_str == Optimiser.SGD:
      self._optimizer = torch.optim.SGD(
          self._q_network.parameters(), lr=learning_rate
      )
    else:
      raise ValueError(
          "Not implemented, choose from 'adam', 'rmsprop', and 'sgd'."
      )

  @torch.inference_mode
  def select_action(
      self, time_step, greedy: bool = True
  ) -> rl_agent.StepOutput:
    # Act step: don't act at terminal info states or if its not our turn.
    """Returns the action to be taken.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      greedy: if the action needs to be greedy, not sampled

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    action = None
    probs = []
    if (not time_step.last()) and (
        time_step.is_simultaneous_move()
        or self.player_id == time_step.current_player()
    ):
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      epsilon = self.epsilon_schedule(self._iteration) if not greedy else 0.0
      action, probs = self._act_epsilon_greedy(
          info_state, legal_actions, epsilon
      )

    return rl_agent.StepOutput(action=action, probs=probs)

  def act_epsilon_greedy(self, info_state, legal_actions, epsilon):
    """Takes an eps-greedy action."""
    return self._act_epsilon_greedy(info_state, legal_actions, epsilon)

  def step(
      self, time_step: rl_environment.TimeStep, is_evaluation: bool = False
  ) -> rl_agent.StepOutput:
    """Returns the action to be taken and updates the Q-network if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: if the agent takes step in an environment (True) or a
        learning step.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    if is_evaluation:
      return self.select_action(time_step, True)

    # Act step: don't act at terminal info states or if its not our turn.
    action, probs = self.select_action(time_step, False)

    self._iteration += 1

    if self._iteration % self._learn_every == 0:
      self._last_loss_value = self.learn()
      if self._last_loss_value != self._last_loss_value:
        return

    if self._iteration % self._update_target_network_every == 0:
      self._copy_weights(self._tau)

    if self._prev_timestep is not None and self._prev_action is not None:
      # We may omit record adding here if it's done elsewhere.
      self.add_transition(self._prev_timestep, self._prev_action, time_step)

    if time_step.last():  # prepare for the next episode.
      self._prev_timestep = None
      self._prev_action = None
      return None
    else:
      self._prev_timestep = time_step
      self._prev_action = action

    return rl_agent.StepOutput(action=action, probs=probs)

  def add_transition(
      self,
      prev_time_step: rl_environment.TimeStep,
      prev_action: int,
      time_step: rl_environment.TimeStep,
  ) -> None:
    """Adds the new transition using `time_step` to the replay buffer.

    Adds the transition from `self._prev_timestep` to `time_step` by
    `self._prev_action`.

    Args:
      prev_time_step: prev `rl_environment.Timestep`
      prev_action: int, action taken at `prev_time_step`.
      time_step: current `rl_environment.Timestep`
    """
    assert prev_time_step is not None
    legal_actions = time_step.observations["legal_actions"][self.player_id]
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=np.asarray(
            prev_time_step.observations["info_state"][self.player_id],
            dtype=np.float32,
        ),
        action=np.asarray(prev_action, dtype=int),
        reward=np.asarray(time_step.rewards[self.player_id], dtype=np.float32),
        next_info_state=np.asarray(
            time_step.observations["info_state"][self.player_id],
            dtype=np.float32,
        ),
        is_final_step=np.asarray(time_step.last(), dtype=bool),
        legal_actions_mask=np.asarray(legal_actions_mask, dtype=bool),
    )
    if self._replay_buffer is None:
      self._replay_buffer = self._replay_buffer_class.init(
          self._replay_buffer_capacity, transition
      )

    self._replay_buffer.append(transition)

  def _act_epsilon_greedy(
      self, info_state: list, legal_actions: list, epsilon: float
  ) -> tuple:
    """Returns a valid epsilon-greedy action and valid action probs.

    Action probabilities are given by a softmax over legal q-values.

    Args:
      info_state (list): hashable representation of the information state.
      legal_actions (list): list of legal actions at `info_state`.
      epsilon (float): probability of taking an exploratory action.

    Returns:
      A valid epsilon-greedy action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions)
    if np.random.rand() < epsilon:
      action = np.random.choice(legal_actions)
      probs[legal_actions] = 1.0 / len(legal_actions)
    else:
      info_state = torch.tensor(
          np.asarray(info_state).reshape(1, -1),
          device=self._device,
          dtype=torch.float32,
      )
      q_values = self._q_network(info_state).detach().squeeze(0)
      legal_q_values = q_values[legal_actions]
      action = legal_actions[torch.argmax(legal_q_values)]
      probs[action] = 1.0

    probs = probs / probs.sum()
    return action, probs

  def learn(self) -> None:
    """Computes the loss on sampled transitions and perform a Q-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Returns:
      The average loss obtained on this batch of transitions or `None`.
    """
    if (
        len(self._replay_buffer) < self._batch_size
        or len(self._replay_buffer) < self._min_buffer_size_to_learn
    ):
      # return None because it's nothing to learn from
      return None

    transitions = self._replay_buffer.sample(self._batch_size)

    info_states = torch.tensor(
        transitions.info_state, device=self._device, dtype=torch.float32
    )
    actions = torch.tensor(
        transitions.action, device=self._device, dtype=torch.long
    )
    rewards = torch.tensor(
        transitions.reward, device=self._device, dtype=torch.float32
    )
    next_info_states = torch.tensor(
        transitions.next_info_state, device=self._device, dtype=torch.float32
    )
    are_final_steps = torch.tensor(
        transitions.is_final_step, device=self._device, dtype=torch.bool
    )
    legal_actions_mask = torch.tensor(
        transitions.legal_actions_mask, device=self._device, dtype=torch.bool
    )

    self._q_values = self._q_network(info_states)
    self._target_q_values = self._target_q_network(next_info_states).detach()

    illegal_actions_mask = torch.logical_not(legal_actions_mask)
    legal_target_q_values = self._target_q_values.masked_fill(
        illegal_actions_mask, ILLEGAL_ACTION_LOGITS_PENALTY
    )
    max_next_q = torch.max(legal_target_q_values, dim=1)[0]

    target = (
        rewards
        + torch.logical_not(are_final_steps)
        * self._discount_factor
        * max_next_q
    ).detach()

    predictions = self._q_values.gather(
        dim=-1, index=actions.unsqueeze(-1)
    ).squeeze(-1)

    loss = self.loss_class(predictions, target)

    self._optimizer.zero_grad()
    loss.backward()

    if self._gradient_norm_clipping is not None:
      nn.utils.clip_grad_norm_(
          self._q_network.parameters(), self._gradient_norm_clipping
      )
    self._optimizer.step()

    return loss.item()

  @property
  def q_values(self) -> torch.Tensor:
    return self._q_values

  @property
  def replay_buffer(self) -> ReplayBuffer:
    return self._replay_buffer

  @property
  def loss(self) -> float:
    return self._last_loss_value

  @property
  def prev_timestep(self) -> rl_environment.TimeStep:
    return self._prev_timestep

  @property
  def prev_action(self) -> np.ndarray:
    return self._prev_action

  @property
  def step_counter(self) -> int:
    return self._iteration

  def _copy_weights(self, tau: float) -> None:
    """Soft update of the target network's weights.

      θ′ ← τ θ + (1 - τ )θ′.

    Args:
        tau (float): main network parameters' weight.
    """
    # Ugly formatting
    for target_network_param, q_network_param in zip(
        self._target_q_network.parameters(), self._q_network.parameters()
    ):
      target_network_param.data.copy_(
          tau * q_network_param.data + (1.0 - tau) * target_network_param.data
      )

  def save(self, data_path: pathlib.Path, save_optimiser: bool = True) -> None:
    """Save checkpoint/trained model and optimizer.

    Args:
      data_path: pathlib.Path for saving model. It can be relative or absolute,
        but filename should be included. For example: q_network.pt or
        /path/to/q_network.pt
      save_optimiser: bool whether to save the optimiser or not. Defaults to
        True.
    """
    checkpoint = {
        "iteration": self._iteration,
        "last_loss_value": self._last_loss_value,
        "model": self._q_network.state_dict(),
    }
    if save_optimiser:
      checkpoint.update({"optimiser": self._optimizer.state_dict()})
    torch.save(checkpoint, data_path)

  def load(self, data_path: pathlib.Path, load_optimiser: bool = True) -> None:
    """Load checkpoint/trained model and optimizer.

    Args:
      data_path: pathlib.Path for saving model. It can be relative or absolute,
        but filename should be included. For example: q_network.pt or
        /path/to/q_network.pt
      load_optimiser: bool, whether to learn the optimiser's state of not.
        Defaults to True.
    """
    checkpoint = torch.load(
        data_path, weights_only=True, map_location=self._device
    )
    self._q_network.load_state_dict(checkpoint["model_state_dict"])
    self._target_q_network.load_state_dict(checkpoint["model_state_dict"])

    if load_optimiser:
      self._optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

    self._iteration = checkpoint["iteration"]
    self._last_loss_value = checkpoint["last_loss_value"]



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Boltzmann DQN agent implemented in JAX.

This algorithm is a variation of DQN that uses a softmax policy directly with
the unregularized action-value function. See https://arxiv.org/abs/2102.01585.
"""

import functools

import chex
import flax.nnx as nn
import jax
import jax.numpy as jnp

from open_spiel.python.jax import dqn


class BoltzmannDQN(dqn.DQN):
  """Boltzmann DQN implementation in JAX."""

  def __init__(self, *args, eta: float = 1.0, **kwargs):
    """Initializes the Boltzmann DQN agent.

    Args:
      *args: args passed to the underlying DQN agent.
      eta: Temperature parameter used in the softmax function.
      **kwargs: kwargs passed to the underlying DQN agent.
    """
    super().__init__(*args, **kwargs)
    self._prev_q_network_state = nn.clone(
        nn.state(self._q_network), variables=True
    )
    self._temperature = eta

  @functools.partial(jax.jit, static_argnums=(0,))
  def _boltzmann_action_probs(
      self,
      network_state: nn.State,
      info_state: chex.Array,
      legal_actions: chex.Array,
      rng: chex.PRNGKey,
      coeff: chex.Array,
  ):
    """Returns a valid soft-max action and action probabilities.

    Args:
      network_state: nn.State of the Q-network.
      info_state: Observations from the environment.
      legal_actions: List of legal actions.
      rng: chex.PRNGKey for randomness
      coeff: If not None, then the terms in softmax function will be
        element-wise multiplied with these coefficients.

    Returns:
      a valid soft-max action and action probabilities.
    """
    q_values = self._jittable_inference(network_state, info_state)
    legal_q_values = jnp.where(
        legal_actions,
        q_values,
        jnp.full_like(q_values, dqn.ILLEGAL_ACTION_LOGITS_PENALTY),
    )
    # Apply temperature and subtract the maximum value for numerical stability.
    temp = legal_q_values / self._temperature
    probs = nn.softmax(coeff * temp)
    action = jax.random.choice(rng, jnp.arange(self._num_actions), p=probs)
    return action, probs

  def _act_epsilon_greedy(
      self,
      network_state: nn.State,
      info_state: chex.Array,
      legal_actions: chex.Array,
      rng: chex.PRNGKey,
      epsilon: float,
  ):
    """Returns a selected action and the probabilities of legal actions."""
    if epsilon == 0.0:  # greeddy evaluation
      # Soft-max normalized by the action probabilities from the previous
      # Q-network.
      prev_rng, rng = jax.random.split(rng)
      _, prev_probs = self._boltzmann_action_probs(
          self._prev_q_network_state,
          info_state,
          legal_actions,
          prev_rng,
          jnp.ones(self._num_actions),
      )
      return self._boltzmann_action_probs(
          network_state, info_state, legal_actions, rng, prev_probs
      )

    # During training, we use the DQN action selection, which will be
    # epsilon-greedy.
    return super()._act_epsilon_greedy(
        network_state, info_state, legal_actions, rng, epsilon
    )

  def update_prev_q_network(self):
    """Updates the parameters of the previous Q-network."""
    self._prev_q_network_state = nn.clone(
        nn.state(self._q_network), variables=True
    )



# Copyright 2026 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the ESCHER agent."""

from absl import app
from absl.testing import absltest
from absl.testing import parameterized

import pyspiel
from open_spiel.python.pytorch import escher


class EscherTest(parameterized.TestCase):

  @parameterized.parameters("kuhn_poker", "leduc_poker")
  def test_escher_runs(self, game_name):
    game = pyspiel.load_game(game_name)

    cfg = escher.Config()
    cfg.value_traversals = 2
    cfg.value_net = [2]
    cfg.value_batch_size = 2
    cfg.value_batch_steps = 2
    cfg.regret_traversals = 2
    cfg.regret_net = [2]
    cfg.regret_batch_size = 2
    cfg.regret_batch_steps = 2
    cfg.avg_policy_net = [2]
    cfg.avg_policy_batch_size = 2
    cfg.avg_policy_batch_steps = 2
    agent = escher.Agent(game, cfg)

    train_cfg = escher.TrainConfig(game)
    train_cfg.iterations = 2
    train_cfg.nashconv = True
    train_cfg.games_vs_random = 1
    escher.train(train_cfg, agent)


def main(_):
  absltest.main()


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A replay buffer for AlphaZero."""

from typing import Generic, TypeVar

import chex
import jax
import jax.numpy as jnp

Experience = TypeVar("Experience", bound=chex.ArrayTree)

"""Educational simplified copy of a flat buffer inspired by
https://github.com/instadeepai/flashbax/blob/main/flashbax/buffers/flat_buffer.py
"""


@chex.dataclass(frozen=True)
class ReplayBufferState(Generic[Experience]):
  experience: Experience
  capacity: chex.Numeric
  entry_index: chex.Array
  is_full: bool


def init(capacity: chex.Numeric, experience: Experience) -> ReplayBufferState:
  """Initialise a replay buffer.

  Args:
    capacity (chex.Numeric, int): max size of the buffer
    experience (Experience): initial value

  Returns:
    ReplayBufferState: state of the buffer
  """
  # Set experience value to be empty.
  experience = jax.tree.map(jnp.empty_like, experience)
  # Broadcast to [add_batch_size, ...]
  experience = jax.tree.map(
      lambda x: jnp.broadcast_to(x[jnp.newaxis, ...], (capacity, *x.shape)),
      experience,
  )
  return ReplayBufferState(
      capacity=capacity,
      experience=experience,
      entry_index=jnp.array(0),
      is_full=jnp.array(False, dtype=jnp.bool),
  )


def append(
    state: ReplayBufferState,
    experience: Experience,
) -> ReplayBufferState:
  """Potentially adds `experience` to the replay buffer.

  Args:
    state: `ReplayBufferState`, current state of the buffer
    experience: data to be added to the reservoir buffer.

  Returns:
    An updated `ReplayBufferState`
  """

  chex.assert_trees_all_equal_dtypes(experience, state.experience)

  index = state.entry_index % state.capacity

  def update_leaf(buffer_leaf: Experience, exp_leaf: Experience):
    return buffer_leaf.at[index].set(exp_leaf)

  new_experience = jax.tree.map(update_leaf, state.experience, experience)

  new_entry_index = state.entry_index + 1
  new_is_full = state.is_full | (new_entry_index >= state.capacity)

  return ReplayBufferState(
      capacity=state.capacity,
      experience=new_experience,
      entry_index=new_entry_index,
      is_full=new_is_full,
  )


def sample(
    rng: chex.PRNGKey, state: ReplayBufferState, num_samples: int
) -> Experience:
  """Returns `num_samples` uniformly sampled from the buffer.

  Args:
    rng: `chex.PRNGKey`, a random state
    state: `ReplayBufferState`, a buffer state
    num_samples: `int`, number of samples to draw.

  Returns:
    An iterable over `num_samples` random elements of the buffer.
  Raises:
    AssertionError: If there are less than `num_samples` elements in the buffer
  """

  # When full, the max time index is max_length_time_axis otherwise it is
  # current index.
  max_size = jnp.where(state.is_full, state.capacity, state.entry_index)
  indices = jax.random.randint(
      rng, shape=(num_samples,), minval=0, maxval=max_size
  )
  return jax.tree.map(lambda x: x[indices], state.experience)


class Buffer:
  """A fixed size buffer that keeps the newest values."""

  def __init__(
      self, max_size: int, force_cpu: bool = False, seed: int = 0
  ) -> None:
    self.max_size = max_size
    self._total_seen = jnp.array(0)

    self._rng = jax.random.PRNGKey(seed)
    # jit-compling all the methods for cpu if forced otherwise automatically
    backend = "cpu" if force_cpu else jax.default_backend()

    self._init = jax.jit(init, backend=backend, static_argnames=("capacity",))
    self._append = jax.jit(append, backend=backend, donate_argnums=(0,))
    self._sample = jax.jit(
        sample, backend=backend, static_argnames=("num_samples",)
    )

    self.buffer_state = None

  @property
  def data(self):
    if self.buffer_state is not None:
      return self.buffer_state.experience

  @property
  def total_seen(self) -> int:
    return self._total_seen.item()

  def __len__(self) -> int:
    if self.buffer_state is None:
      return 0
    return self.max_size if self.buffer_state.is_full else self._total_seen

  def __bool__(self) -> bool:
    return self.buffer_state is not None

  def append(self, val: Experience) -> None:
    if self.buffer_state is None:
      self.buffer_state = self._init(self.max_size, val)
    self.buffer_state = self._append(self.buffer_state, val)
    self._total_seen = self.buffer_state.entry_index

  def shuffle(self, key: int) -> chex.Array:
    return jax.random.permutation(jax.random.key(key), jnp.arange(len(self)))

  def sample(self, num_samples: int) -> Experience:
    self._rng, rng = jax.random.split(self._rng)
    batch = self._sample(rng, self.buffer_state, num_samples)
    return batch



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Example usage of the minimax solver."""

from absl import app
from absl import flags

from open_spiel.python.algorithms import minimax_solver
import pyspiel

FLAGS = flags.FLAGS

# Game strings can just contain the name or the name followed by parameters
# and arguments, e.g. "breakthrough(rows=6,columns=6)"
flags.DEFINE_string("game_string", "tic_tac_toe", "Game string")


def main(_):
  print("Creating game.")
  game = pyspiel.load_game(FLAGS.game_string)
  state = game.new_initial_state()
  print(state)

  print("Creating minimax oracle and solving.")
  solver = minimax_solver.MinimaxSolver(FLAGS.game_string)
  solver.solve()

  print("Playing game.")
  while not state.is_terminal():
    print("")
    print(state)
    # Decision node: sample action for the single current player
    action_values = solver.action_values_from_state(state)
    best_value = float("-inf")
    best_action = pyspiel.INVALID_ACTION
    for action in state.legal_actions():
      action_value = action_values[action]
      print(
          f"Action {state.action_to_string(action)} "
          + f"has minimax value: {action_value}"
      )
      if action_value > best_value:
        best_value = action_value
        best_action = action
    print(
        f"Applying action {best_action}: "
        + f"{state.action_to_string(best_action)}"
    )
    state.apply_action(best_action)

  # Game is now done. Print utilities for each player
  print("")
  print(str(state))
  returns = state.returns()
  for pid in range(game.num_players()):
    print("Utility for player {} is {}".format(pid, returns[pid]))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Population Exploitability Descent (PED) Stochastic Approx. Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.nonsymmetric import updates


class Solver(updates.Solver):
  """PED Solver."""

  def __init__(self, proj_grad=True, euclidean=False, lrs=(1e-1,),
               rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    super().__init__(proj_grad, euclidean, rnd_init, seed)
    self.lrs = lrs

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return exploitability.

    Args:
      params: tuple of params (dist,), see ped.gradients
      payoff_matrices: dictionary with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    Returns:
      float, exploitability of current dist
      unregularized exploitability (stochastic estimate)
      unregularized exploitability (stochastic estimate) *duplicate
    """
    return gradients(*params, payoff_matrices, self.num_players, self.proj_grad)


def gradients(dist, payoff_matrices, num_players, proj_grad=True):
  """Computes exploitablity gradient.

  Args:
    dist: list of 1-d np.arrays, current estimate of nash distribution
    payoff_matrices: dictionary with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    num_players: int, number of players, in case payoff_matrices is abbreviated
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    unregularized exploitability (stochastic estimate) *duplicate
  """
  # first compute best responses and payoff gradients
  nabla = []
  br = []
  unreg_exp = []
  for i in range(num_players):

    nabla_i = np.zeros_like(dist[i])
    for j in range(num_players):
      if j == i:
        continue
      if i < j:
        hess_i_ij = payoff_matrices[(i, j)][0]
      else:
        hess_i_ij = payoff_matrices[(j, i)][1].T

      nabla_ij = hess_i_ij.dot(dist[j])
      nabla_i += nabla_ij / float(num_players - 1)

    nabla.append(nabla_i)

    power = np.inf
    s_i = np.linalg.norm(nabla_i, ord=power)
    br_i = np.zeros_like(nabla_i)
    maxima_i = (nabla_i == s_i)
    br_i[maxima_i] = 1. / maxima_i.sum()
    br.append(br_i)

    unreg_exp.append(np.max(nabla_i) - nabla_i.dot(dist[i]))

  # then construct exploitability gradient
  grad_dist = []
  for i in range(num_players):

    grad_dist_i = -nabla[i]
    for j in range(num_players):
      if j == i:
        continue
      if i < j:
        hess_j_ij = payoff_matrices[(i, j)][1]
      else:
        hess_j_ij = payoff_matrices[(j, i)][0].T

      grad_dist_i += hess_j_ij.dot(br[j] - dist[j])

    if proj_grad:
      grad_dist_i = simplex.project_grad(grad_dist_i)

    grad_dist.append(grad_dist_i)

  return (grad_dist,), np.mean(unreg_exp), np.mean(unreg_exp)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A model based on nnx."""

import os
from typing import Callable, Optional, Sequence
import warnings

import chex
import flax.nnx as nn
from flax.training import train_state
import jax
import jax.numpy as jnp
import numpy as np
import optax
import orbax.checkpoint as orbax

# pylint: disable=g-bare-generic

from open_spiel.python.algorithms.alpha_zero import utils

warnings.warn("Pay attention that you've been using the `nnx` api")

activations_dict = {
    "celu": nn.celu,
    "elu": nn.elu,
    "gelu": nn.gelu,
    "glu": nn.glu,
    "hard_sigmoid": nn.hard_sigmoid,
    "hard_silu": nn.hard_silu,  # Alias for hard_swish
    "hard_swish": nn.hard_swish,  # Alias for hard_silu
    "hard_tanh": nn.hard_tanh,
    "leaky_relu": nn.leaky_relu,
    "log_sigmoid": nn.log_sigmoid,
    "log_softmax": nn.log_softmax,
    "relu": nn.relu,
    "selu": nn.selu,
    "sigmoid": nn.sigmoid,
    "silu": nn.silu,
    "soft_sign": nn.soft_sign,
    "softmax": nn.softmax,
    "softplus": nn.softplus,
    "swish": nn.swish,
    "tanh": nn.tanh,
}


def get_batch_stats(layer: nn.Module) -> nn.BatchStat:
  """Filter to get the model's parameters."""

  return nn.state(layer, nn.BatchStat)


def get_layer_parameters(layer: nn.Module) -> nn.Param:
  """Filter to get the model's BatchNorm stats."""

  return nn.state(layer, nn.Param)


class Flatten(nn.Module):
  """A simple `nn.Module` class wrapper for the flattening the output."""

  def __init__(self) -> None:
    self.layer = utils.flatten

  def __call__(self, x: chex.Array) -> chex.Array:
    return self.layer(x)


class Activation(nn.Module):
  """A simple `nn.Module` class wrapper for the activations."""

  def __init__(self, activation_name: str) -> None:
    """Intialises the activation function.

    Args:
        activation_name (str): a name of the activation.
        The following are available: {", ".join(list(activations_dict.keys()))}
    """
    self.activation_name = activation_name

  def __call__(self, x: chex.Array) -> chex.Array:
    # if the activation not found or None just passes the identity function
    return activations_dict.get(self.activation_name, lambda x: x)(x)


class MLPBlock(nn.Module):
  """Multi layer perceptron block with a Linear layer + Activation."""

  def __init__(
      self, in_features: int, out_features: int, activation: str, seed: int = 0
  ) -> None:
    self.activation = Activation(activation)
    self.dense_layer = nn.Linear(in_features, out_features, rngs=nn.Rngs(seed))

  def __call__(self, x: chex.Array) -> chex.Array:
    y = self.dense_layer(x)
    y = self.activation(y)
    return y


class ConvBlock(nn.Module):
  """Convolutional block with SAME padding + BatchNorm + Activation."""

  def __init__(
      self,
      in_features: int,
      out_features: int,
      kernel_size: tuple[int, int],
      activation: str,
      seed: int = 0,
  ) -> None:
    self.conv = nn.Conv(
        in_features,
        out_features,
        kernel_size=kernel_size,
        padding="SAME",
        rngs=nn.Rngs(seed),
    )
    self.activation = Activation(activation)
    self.bn = nn.BatchNorm(out_features, rngs=nn.Rngs(seed), axis_name="batch")

  def output_shape(self, x: chex.Array) -> tuple[int, int, int]:
    # NOTE: for the given settings: padding="SAME" and strides=(1,1)
    # output size is equal to the input size, otherwise the layer should
    # implement this method. For more, see:
    # https://blog.mlreview.com/a-guide-to-receptive-field-arithmetic-for-convolutional-neural-networks-e0f514068807
    return x.shape

  def __call__(self, x: chex.Array) -> chex.Array:
    y = self.conv(x)
    y = self.bn(y)
    y = self.activation(y)
    return y


class ResidualBlock(nn.Module):
  """Residual block with two conv blocks and an addition."""

  def __init__(
      self,
      in_features: int,
      out_features: int,
      kernel_size: tuple[int, int],
      activation: str,
      seed: int = 0,
  ) -> None:
    self.conv1 = ConvBlock(
        in_features, out_features, kernel_size, activation, seed
    )
    # activation's applied separately
    self.conv2 = ConvBlock(out_features, out_features, kernel_size, None, seed)
    self.activation = Activation(activation)

  def output_shape(self, x: chex.Array) -> tuple[int, int, int]:
    return self.conv2.output_shape(x)

  def __call__(self, x: chex.Array) -> chex.Array:
    residual = x
    y = self.conv1(x)
    y = self.conv2(y)
    y = self.activation(y + residual)
    return y


class PolicyHead(nn.Module):
  """Policy network."""

  def __init__(
      self,
      input_shape: chex.Shape,
      nn_width: int,
      out_features: int,
      model_type: str,
      activation: str,
      seed: int = 0,
  ) -> None:
    *space_features, in_features = input_shape

    if model_type == "mlp":
      self.torso = MLPBlock(in_features, nn_width, activation, seed)
    else:
      self.torso = nn.Sequential(
          ConvBlock(in_features, 2, (1, 1), activation, seed),
          Flatten(),
          MLPBlock(np.prod(space_features) * 2, nn_width, activation, seed),
      )

    self.policy_head = MLPBlock(nn_width, out_features, None, seed)

  def __call__(self, x: chex.Array) -> chex.Array:
    y = self.torso(x)
    policy_logits = self.policy_head(y)
    return policy_logits


class ValueHead(nn.Module):
  """Value network."""

  def __init__(
      self,
      input_shape: chex.Shape,
      nn_width: str,
      model_type: str,
      activation: str,
      seed: int,
  ) -> None:
    *space_features, in_features = input_shape

    if model_type == "mlp":
      self.torso = MLPBlock(in_features, nn_width, activation, seed)
    else:
      self.torso = nn.Sequential(
          ConvBlock(in_features, 1, (1, 1), activation),
          Flatten(),
          MLPBlock(np.prod(space_features) * 1, nn_width, activation, seed),
      )

    self.value_head = MLPBlock(nn_width, 1, "tanh", seed)

  def __call__(self, x: chex.Array) -> chex.Array:
    y = self.torso(x)
    values = self.value_head(y)
    return values


class AlphaZeroModel(nn.Module):
  """An AlphaZero style model with a policy and value head.

  This supports three types of models: mlp, conv2d and resnet.

  All models have a shared torso stack with two output heads: policy and value.
  They have same meaning as in the AlphaGo Zero and AlphaZero papers. The resnet
  model copies the one in that paper when set with width 256 and depth 20. The
  conv2d model is the same as the resnet except uses a conv+batchnorm+relu
  instead of the res blocks. The mlp model uses dense layers instead of conv,
  and drops batch norm.

  Links to relevant articles/papers:
    https://deepmind.com/blog/article/alphago-zero-starting-scratch has an open
      access link to the AlphaGo Zero nature paper.
    https://deepmind.com/blog/article/alphazero-shedding-new-light-grand-games-chess-shogi-and-go
      has an open access link to the AlphaZero science paper.

  All are parameterized by their input (observation) shape and output size
  (number of actions), though the conv2d and resnet might only work with games
  that have spatial data (ie 3 non-batch dimensions, eg: connect four would
  work, but not poker).

  The depth is the number of blocks in the torso, where the definition of a
  block varies by model. For a resnet it's a resblock which is two conv2ds,
  batch norms and relus, and an addition. For conv2d it's a conv2d, a batch norm
  and a relu. For mlp it's a dense plus relu.

  The width is the number of filters for any conv2d and the number of hidden
  units for any dense layer.
  """

  def __init__(
      self,
      model_type: str,
      input_shape: chex.Shape,
      output_size: int,
      nn_width: int,
      nn_depth: int,
      activation: Optional[str] = "relu",
      seed: Optional[int] = 0,
  ) -> None:
    """An AlphaZero model with a policy and value heads.

    Args:
      model_type (str): underlying model type: conv2d, mlp, or resnet
      input_shape (chex.Shape): input spec
      output_size (int): action space spec
      nn_width (int): hidden layers dimensionality
      nn_depth (int): number of hidden layers
      activation (str): an activation for the neural networks. Defaults to
        "relu"
      seed (int): random seed for the network
    """

    # torso
    if model_type == "mlp":
      self.torso = nn.Sequential(
          Flatten(),
          MLPBlock(
              np.prod(input_shape), nn_width, activation=activation, seed=seed
          ),
          *[
              MLPBlock(
                  nn_width, nn_width, activation=activation, seed=seed + sdo
              )
              for sdo in range(1, nn_depth)
          ],
      )
    elif model_type == "conv2d":
      self.torso = nn.Sequential(
          # by default an observation may be flat
          lambda x: x.reshape(input_shape),
          ConvBlock(input_shape[-1], nn_width, (3, 3), activation, seed=seed),
          *[
              ConvBlock(
                  nn_width, nn_width, (3, 3), activation, seed=seed + sdo
              )
              for sdo in range(nn_depth)
          ],
      )
    elif model_type == "resnet":
      self.torso = nn.Sequential(
          lambda x: x.reshape(input_shape),
          ConvBlock(input_shape[-1], nn_width, (3, 3), activation, seed=seed),
          *[
              ResidualBlock(
                  nn_width, nn_width, (3, 3), activation, seed=seed + sdo
              )
              for sdo in range(nn_depth)
          ],
      )
    else:
      raise ValueError(f"Unknown model type: {model_type}")

    self.policy_head = PolicyHead(
        (*input_shape[:-1], nn_width),
        nn_width,
        output_size,
        model_type,
        activation,
        seed,
    )
    self.value_head = ValueHead(
        (*input_shape[:-1], nn_width), nn_width, model_type, activation, seed
    )

  def __call__(self, observations: chex.Array) -> tuple[chex.Array, chex.Array]:
    x = self.torso(observations)
    policy_logits = self.policy_head(x)
    value_out = self.value_head(x)

    return policy_logits, value_out.squeeze(-1)


# modified train state
class TrainState(train_state.TrainState):
  batch_stats: nn.State
  graphdef: nn.GraphDef


class Model:
  """Basic model class."""
  valid_model_types = ["mlp", "conv2d", "resnet"]

  def __init__(
      self, state: TrainState, path: str, update_step_fn: Callable
  ) -> None:
    self._state = state
    self._path = path
    self._update_step_fn = update_step_fn

    # we're using the latest checkpointing API, see:
    # https://flax.readthedocs.io/en/latest/guides/checkpointing.html

    self._checkpointer = None

    if self._path is not None:
      self._checkpointer = orbax.Checkpointer(orbax.StandardCheckpointHandler())

  @classmethod
  def _create_train_state(
      cls, model: AlphaZeroModel, optimiser: optax.GradientTransformation
  ) -> TrainState:
    graphdef, variables, batch_stats = nn.split(model, nn.Param, nn.BatchStat)

    return TrainState.create(
        apply_fn=model,
        params=variables,
        tx=optimiser,
        batch_stats=batch_stats,
        graphdef=graphdef,
    )

  @classmethod
  def build_model(
      cls,
      model_type: str,
      input_shape: chex.Shape,
      output_size: chex.Numeric,
      nn_width: int,
      nn_depth: int,
      weight_decay: float,
      learning_rate: float,
      path: str,
      seed: int = 0,
      decouple_weight_decay: bool = False,
  ) -> "Model":
    """Builds the model."""

    if model_type not in cls.valid_model_types:
      raise ValueError(
          f"Invalid model type: {model_type}, expected one of:"
          f" {cls.valid_model_types}"
      )

    if isinstance(input_shape, int):
      input_shape = (input_shape,)
    elif isinstance(input_shape, list):
      input_shape = tuple(input_shape)

    while len(input_shape) < 3:
      input_shape = input_shape + (1,)

    model = AlphaZeroModel(
        model_type=model_type,
        input_shape=input_shape,
        output_size=output_size,
        nn_width=nn_width,
        nn_depth=nn_depth,
        seed=seed,
    )

    def mask_fn(path, _):
      # path is a tuple of segments, e.g., ('linear1', 'bias')
      names = [
          str(p.key) if isinstance(p, jax.tree_util.DictKey) else str(p)
          for p in path
      ]
      # Return True to APPLY decay, False to MASK it
      return ("bias" not in names) and ("bn" not in names)

    params = nn.state(model, nn.Param)
    mask = jax.tree.map_with_path(mask_fn, params)

    weights_no_bias_filter = nn.All(nn.Param, mask_fn)

    if not decouple_weight_decay:
      optimiser = optax.adam(learning_rate=learning_rate)
    else:
      optimiser = optax.adamw(
          learning_rate=learning_rate, weight_decay=weight_decay, mask=mask
      )

    state = cls._create_train_state(model, optimiser)

    @nn.vmap(in_axes=(None, 0), axis_name="batch")
    def forward(model: AlphaZeroModel, x: chex.Array) -> chex.Array:
      return model(x)

    @jax.jit
    def update_step_fn(
        state: TrainState,
        observations: chex.Array,
        legals_mask: chex.Array,
        policy_targets: chex.Array,
        value_targets: chex.Array,
    ) -> Callable:
      def loss_fn(
          model,
          observations: chex.Array,
          legals_mask: chex.Array,
          policy_targets: chex.Array,
          value_targets: chex.Array,
      ):
        policy_logits, value_preds = forward(model, observations)
        policy_logits = jnp.where(
            legals_mask,
            policy_logits,
            jnp.full_like(policy_logits, jnp.finfo(jnp.float32).min),
        )

        policy_loss = optax.softmax_cross_entropy(
            policy_logits, policy_targets
        ).mean()
        value_loss = optax.l2_loss(value_preds, value_targets).mean()

        l2_reg_loss = (
            optax.tree_utils.tree_norm(
                nn.state(model, weights_no_bias_filter), ord=2, squared=True
            )
            * weight_decay
        )

        total_loss = policy_loss + value_loss

        if not decouple_weight_decay:
          total_loss = total_loss + l2_reg_loss

        return total_loss, (
            policy_loss,
            value_loss,
            l2_reg_loss,
            get_batch_stats(model),
        )

      grad_fn = nn.value_and_grad(loss_fn, has_aux=True)
      model = nn.merge(state.graphdef, state.params, state.batch_stats)
      model.train()

      (_, (policy_loss, value_loss, l2_reg_loss, new_batch_stats)), grads = (
          grad_fn(
              model,
              observations,
              legals_mask,
              policy_targets,
              value_targets,
          )
      )

      new_state = state.apply_gradients(
          grads=grads, batch_stats=new_batch_stats
      )

      return new_state, (policy_loss, value_loss, l2_reg_loss)

    return cls(state, path, update_step_fn)

  @classmethod
  def from_checkpoint(cls, checkpoint_path, model_args=None):
    if model_args is None:
      raise ValueError("model_args must be provided for checkpoint loading")

    model = cls.build_model(**model_args)
    model._state = cls.load_checkpoint(checkpoint_path)

    return model

  @property
  def num_trainable_variables(self) -> int:
    return sum(np.prod(p.shape) for p in jax.tree.leaves(self._state.params))

  def print_trainable_variables(self) -> None:
    model = nn.merge(
        self._state.graphdef, self._state.params, self._state.batch_stats
    )
    nn.display(model)
    flat_params, _ = jax.tree.flatten(self._state.params)
    for i, p in enumerate(flat_params):
      print(f"Param {i}: {p.shape}")

  def inference(
      self, observation: chex.Array, legals_mask: chex.Array
  ) -> tuple[chex.Array, chex.Array]:
    """Returns the policy and value for a given observation."""
    model = nn.merge(
        self._state.graphdef, self._state.params, self._state.batch_stats
    )
    model.eval()

    observation = jnp.asarray(observation, dtype=jnp.float32)
    legals_mask = jnp.asarray(legals_mask, dtype=jnp.bool)

    policy_logits, value = model(observation)
    policy_logits = jnp.where(
        legals_mask,
        policy_logits,
        jnp.full_like(policy_logits, jnp.finfo(jnp.float32).min),
    )

    policy = nn.softmax(policy_logits)
    return value, policy

  def update(
      self,
      batch: Sequence[utils.TrainInput] | chex.ArrayTree
  ) -> utils.Losses:
    """Updates the model with a batch of data."""
    self._state, (policy_loss, value_loss, l2_reg_loss) = self._update_step_fn(
        self._state,
        batch.observation,
        batch.legals_mask,
        batch.policy,
        batch.value,
    )

    return utils.Losses(policy=policy_loss, value=value_loss, l2=l2_reg_loss)

  def save_checkpoint(self, step: int) -> int:
    """Saves a checkpoint of the model."""
    path = os.path.join(self._path, f"checkpoint-{step}")
    if self._checkpointer:
      self._checkpointer.save(path, self._state, force=True)
    else:
      print("No checkpoint path is provided. Skipping.")
    return step

  def load_checkpoint(self, step: int) -> None:
    """Loads a checkpoint of the model."""
    if self._checkpointer:
      self._state = self._checkpointer.restore(
          os.path.join(self._path, f"checkpoint-{step}"), self._state
      )
